package Mooix::MCP;
use strict;
use warnings;
my $quoted_string=qr/"(?:[^"]|\\")*"/;
my $unquoted_string=qr/[-~`!@#$%^&)(=+}{\]\[|';\/><.,A-Za-z0-9_]+/;
my $ident=qr/[a-zA-Z_][-a-zA-Z_]*/;
sub new {
	my $proto=shift;
	my $class=ref($proto) || $proto;
	my $this=bless({thing => shift(), sep => shift(), pending => {}}, 
		       $class);
	return $this;
}
sub escape {
	my $this=shift;
	my $line=shift;
	if ($line=~/#\$(?:"|#)/) {
		return '#$"'.$line;
	}
	else {
		return $line;
	}
}
sub decode {
	my $this=shift;
	my $line=shift;
	$line=~s/$this->{sep}$//;
	
	if ($line=~/^#\$"/) {
		$line=~s/^#\$"//;
		return $line;
	}
	elsif ($line=~/^#\$#/) {
		if ($line=~/^#\$#($ident)\s+($unquoted_string)(?:$|\s+(.*))/) {
			my $message=$1;
			my $auth_key=$2;
			my $keyvals=$3;
			if ($auth_key ne $this->auth_key) {
				return;
			}
			$this->handle_message($message, $keyvals)
		}
		elsif ($line=~/^#\$#\*\s+($unquoted_string)\s+($ident): (.*)/) {
			my $data_tag=$1;
			my $data_key=$2;
			my $data_value=$3;
			$this->add_value($data_tag, $data_key, $data_value);
		}
		elsif ($line=~/^#\$#:\s+($unquoted_string)\s*$/) {
			my $data_tag=$1;
			$this->send_pending_message($data_tag);
		}
		elsif ($line=~/^#\$#($ident)\s+(.*)/) {
			my $message=$1;
			my $keyvals=$2;
			if (length $this->auth_key) {
				return;
			}
			$this->handle_message($message, $keyvals);
		}
		return;
	}
	else {
		return $line;
	}
}
sub encode {
	my $this=shift;
	my $message=shift;
	
	my %params;
	my @keylist;
	while (@_) {
		my $key=shift;
		my $value=shift;
		push @keylist, $key unless exists $params{$key};
		push @{$params{$key}}, $value;
	}
	my $s="#\$#$message";
	if (length $this->auth_key) {
		$s.=" ".$this->auth_key;
	}
	my $e="";
	my $data_tag='';
	foreach my $key (@keylist) {
		if (@{$params{$key}} > 1 || $key =~ /\*$/) {
			my $keyname=$key;
			$keyname=~s/\*$//;
			$s.=" $keyname\*: \"\"";
			if (! length $data_tag) {
				$data_tag=int(rand(1000000));
				$s.=" _data-tag: $data_tag";
			}
			foreach (@{$params{$key}}) {
				$e.="#\$#* $data_tag $keyname\: ".$_.$this->{sep};
			}
		}
		else {
			my $value=${$params{$key}}[0];
			if ($value !~ /^$unquoted_string$/) {
				$value=~s/\"/\\"/g;
				$value='"'.$value.'"';
			}
			$s.=" $key: $value";
		}
	}
	if (length $data_tag) {
		$e.='#$#: '.$data_tag.$this->{sep};
	}
	return $s.$this->{sep}.$e;
}
sub auth_key {
	my $this=shift;
	
	if (exists $this->{auth_key}) {
		return $this->{auth_key};
	}
	else {
		my $k=$this->{thing}->mcpsession_auth_key;
		if (defined $k && length $k) {
			$this->{auth_key}=$k;
			return $k;
		}
	}
	return "";
}
sub known_data_tag {
	my $this=shift;
	my $tag=shift;
	foreach my $t (@{$this->{data_tags}}) {
		return 1 if $tag eq $t;
	}
	return 0;
}
sub handle_message {
	my $this=shift;
	my $message=shift;
	my $keyvals=shift;
	
	my (@keyvals)=$keyvals=~/\s*($ident)(\*?):\s+($unquoted_string|$quoted_string)\s*/g;
	my %params;
	my %multilineparams;
	while (@keyvals) {
		my $key=shift @keyvals;
		my $multiline=shift @keyvals;
		my $value=shift @keyvals;
		if ($value=~/^"/) {
			$value=~s/^"//;
			$value=~s/"$//;
			$value=~s/\\"/"/g;
		}
		if ($multiline eq '*') {
			$multilineparams{$key}=[];
		}
		else {
			return if exists $params{$key};
			$params{$key}=$value;
		}
	}
	my $callback=sub {
		if (@_) {
			my $key=shift;
			my $value=shift;
			return unless $multilineparams{$key};
			push @{$multilineparams{$key}}, $value;
		}
		else {
			my @params;
			foreach my $key (keys %params) {
				push @params, $key, $params{$key};
			}
			foreach my $key (keys %multilineparams) {
				foreach my $value (@{$multilineparams{$key}}) {
					push @params, $key, $value;
				}
			}
			$this->{thing}->mcpsession_receive($message, @params);
		}
	};
	if (keys %multilineparams) {
		return unless exists $params{'_data-tag'} && length $params{'_data-tag'};
		$this->{pending}->{$params{'_data-tag'}}=$callback;
		delete $params{'_data-tag'};
	}
	else {
		$callback->();
	}
}
sub add_value {
	my $this=shift;
	my $tag=shift;
	my $key=shift;
	my $value=shift;
	return unless ref $this->{pending}->{$tag};
	$this->{pending}->{$tag}->($key, $value);
}
sub send_pending_message {
	my $this=shift;
	my $tag=shift;
	return unless ref $this->{pending}->{$tag};
	$this->{pending}->{$tag}->();
	delete $this->{pending}->{$tag};
}
1
