package Mooix::Thing;
use constant {
	_USEPATH	=> 0,
	_ABSPATH	=> 1,
	_INDEX		=> 2,
	_EXEC_NEXT 	=> 3,
	_LAST_FIELD	=> 4
};
our $debugging = 0;
sub import {
	my $pkg = shift;
	my $callpkg = caller;
	
	
	if (exists $ENV{METHOD}) {
		$0=$ENV{METHOD};
	}
	*debuglog = *_debuglog_null;
	*{"${callpkg}::fail"} = sub {
		print join("\n",map { "\"$_\"" } @_)."\n";
		exit 10;
	};
	*{"${callpkg}::run"} = sub {
		if (! exists $ENV{THIS}) {
			die "\$ENV{THIS} is not set. Not running under mooix?\n";
		}
		
		chdir($ENV{THIS});
		my $this = Mooix::Thing->get($ENV{THIS});
		$this->[_USEPATH] = '.';
		if (exists $ENV{MOOIX_DEBUG}) {
			$this->debugger($ENV{MOOIX_DEBUG});
		}
		
		my @params=$this;
		if (prototype($_[0]) ne '$') {
			push @params, map {
				chomp;
				if (/^mooix:/) {
					$_=Mooix::Thing->get($_);
				}
				elsif (/^"(.*)"$/) {
					$_=$1;
					s/(?<!\\)\\n/\n/g;
					s/\\\\/\\/g;
				}
				$_
			} <STDIN>;
		}
		for ($_[0]->(@params)) {
			if (! ref $_) {
				my $f=$_;
				$f =~ s/\\/\\\\/g;
				$f =~ s/\n/\\n/g;
				print "\"$f\"\n";
			}
			else {
				print $_, "\n";
			}
		}
	}
}
sub get {
	my $proto=shift;
	my $id=shift;
	
	$id=~s/^mooix://;
	return unless -e "$id/.mooix";
	my $abspath=$id;
	if (-l $abspath) {
		$abspath=readlink($id);
		if ($abspath !~ m:^/:) {
			print STDERR "warning: $abspath of $id is not absolute.\n";
			my ($dir)=$id=~m!(.*)/[^/]!;
			$abspath="$dir/$abspath";
		}
	}
	elsif ($abspath !~ m:^/:) {
		my $pwd;
		if (@_ && $_[0] =~ m:^/:) {
			$pwd=$_[0];
		}
		else {
			$pwd=`pwd`;
			chomp $pwd;
		}
		$abspath="$pwd/$abspath";
	}
	
	my @s=stat("$id/.");
	bless([$abspath, $abspath, "$s[0].$s[1]"], ref($proto) || $proto);
}
sub id {
	$_[0]->[_ABSPATH];
}
sub index {
	$_[0]->[_INDEX];
}
sub usage {
	my $this=shift;
	my ($caller, $method) = $0 =~ m!^(.*)/(.*)!;
	
	$method = $0 if $method eq '';
	die join("\n", @_, "Usage: ", $this->getusage(field => $method));
}
sub fieldfile {
	my $this=shift;
	my $field=shift;
	$field=~s/^_/./;
	
	if (-l "$$this[0]/$field" || -e _) {
		return "$$this[0]/$field";
	}
	elsif ($field ne 'parent') {
		my $dir="$$this[0]/parent";
		my $c=0;
		while (-e "$dir/.mooix") {
			lstat("$dir/$field");
			return "$dir/$field" if -e _;
			$dir.="/parent";
			if ($c++ > 200) {
				die "possible recursive parent loop: $dir";
			}
		}
	}
	if ($field =~ /_.+/) {
		my ($mixin, $field)=split(/_/, $field, 2);
		my $dir=$this->[_USEPATH];
		my $c=0;
		do {
			if (-e "$dir/$mixin/.mooix") {
				if (-l "$dir/$mixin/$field" || -e _) {
					return "$dir/$mixin/$field";
				}
				$dir.="/$mixin/parent";
				my $c2=0;
				while (-e "$dir/.mooix") {
					lstat("$dir/$field");
					return "$dir/$field" if -e _;
					$dir.="/parent";
					if ($c2++ > 200) {
						die "possible recursive parent loop: $dir";
					}
				}
				
				return undef;
			}
			$dir.="/parent";
			if ($c++ > 200) {
				die "possible recursive parent loop: $dir";
			}
		} until (! -e "$dir/.mooix");
	}
	return undef;
}
sub super {
	my $this=shift;
	
	my ($caller, $method) = $0 =~ m!^(.*)/(.*)!;
	$method = $0 if $method eq '';
	
	my $prefix="$$this[0]/$caller/parent";
	while (-e "$prefix/.mooix") {
		lstat("$prefix/$method");
		if (-f _) {
			if (-x _) {
				$this->debuglog(type => "call", field => "$prefix/$method", message => \@_) if $debugging;
				return $this->_runmethod("$prefix/$method", @_);
			}
			else {
				if (@_) {
					$this->croak("super is a field");
				}
				else {
					$this->debuglog(type => "read", field => $method) if $debugging;
					return $this->_readfield("$prefix/$method");
				}
			}
		}
		$prefix.="/parent";
	}
}
sub exec {
	$_[0]->[_EXEC_NEXT] = 1;
	$_[0];
}
sub getlock {
	my $this=shift;
	my $locktype=shift;
	my $file=shift;
	
	$file='.mooix' if ! defined $file;
	my $lock;
	open ($lock, "$$this[0]/$file") || return undef;
	flock ($lock, $locktype) || return undef;
	$this->debuglog(type => "lock", field => $file, message => "type: $locktype") if $debugging;
	
	return $lock;
}
sub isa {
	my $this=shift;
	my $class=shift;
	if (! ref $class) {
		return UNIVERSAL::isa($this, $class);
	}
	
	while (ref $this) {
		return 1 if ($this == $class);
		$this = $this->parent;
	}
	return 0;
}
sub msg {
	my $this=shift;
	$this->debuglog(type => "call", field => "msg", message => ["event", @_]) if $debugging;
	$this->_runmethod($this->fieldfile("msg"), event => @_);
}
sub croak {
	my $this=shift;
	require Carp;
	my $message="Error: ".Carp::shortmess(shift().(length $! ? " ($!)" : ""));
	my ($file) = $0 =~ /.*\/+(.*)/;
	$obj = Mooix::Thing->get($ENV{THIS});
	$message=~s/\Q$0\E (line \d+)/$obj->$file $1/;
	chomp $message;
	require Mooix::Error;
	die Mooix::Error->new($this, $message);
}
sub _debuglog_null { 0 }
my ($debugobj, $actingobj, $methodfile, $method, @debug_wanted);
sub _debuglog_real {
	my $this=shift;
	
	return unless ref $debugobj && $debugging;
	my %params=@_;
	if (exists $params{field}) {
		$params{field}=~s/.*\///;
	}
	my $expr=$params{type}." ".(exists $params{field} ? $params{field} : "");
	return unless grep { $expr=~/^$_$/ } @debug_wanted;
	
	my $savedstatuscode=$?;
	$debugging=0;
	my @params;
	if (ref $params{message} eq 'ARRAY') {
		foreach my $param (@{$params{message}}) {
			push @params, message => $param;
		}
		delete $params{message};
	}
	elsif (ref $params{message} eq 'CODE') {
		$params{message}=$params{message}->();
	}
	my $value;
	if ($params{replace}) {
		foreach my $param (@{$params{value}}) {
			push @params, value => $param;
		}
		$value=delete $params{value};
	}
	my @ret = $debugobj->_runmethod(
		$methodfile, %params, @params,
		object => $actingobj,
		method => $method,
		accessee => $this
	);
	if ($params{replace}) {
		@{$value} = @ret;
	}
	else {
		@debug_wanted=map {
			s/(^| )\*( |$)/$1.*$2/g;
			s/[^A-Za-z0-9_.* ]//g;
			$_
		} @ret;
	}
	$debugging = @debug_wanted;
	
	$?=$savedstatuscode;
	return $ret;
}
sub debugger {
	my $this=shift;
	if (@_) {
		my $debugger=shift;
		if (! $debugger) {
			delete $ENV{MOOIX_DEBUG};
			$debugging=0;
			*debuglog = *_debuglog_null;
		}
		else {
			$ENV{MOOIX_DEBUG}=$debugger;
			$debugging=1;
			*debuglog = *_debuglog_real;
			$debugobj = Mooix::Thing->get($debugger);
			if (! ref $debugobj) {
				$this->croak("bad debug object, $debugger");
			}
			$actingobj = Mooix::Thing->get($ENV{THIS});
			$method=$0;
			$method=~s/^.*\///;
			$methodfile = $debugobj->fieldfile("log");
			@debug_wanted = '.*';
		}
	}
	if ($debugging) {
		return $this->get($ENV{MOOIX_DEBUG});
	}
	else {
		return;
	}
}
sub implements {
	my $this=shift;
	my $can=$this->can(@_);
	return $can if $can;
	my $file=$this->fieldfile(@_);
	return $file if length $file && -x $file && ! -d _;
}
sub defines {
	my $file=$_[0]->[_ABSPATH]."/$_[1]";
	return $file if -e $file;
}
sub encapsulator {
	my $this=shift;
	return $this->get($this->[_ABSPATH]."/..");
}
sub safegetfield {
	my $this = shift;
	my $field = shift;
	return if $field=~/^[_.]/;
	return if $this->can($field);
	my $file=$this->fieldfile($field);
	return if defined $file && -d _ || -k _;
	if (defined $file && -f _ && -x _) {
		my $safefield=".$field-safe";
		return unless $this->$safefield;
        }
	
	return eval {$this->$field};
}
sub background {
	return 0 if fork;
	close STDIN;
	close STDOUT;
	open STDIN, "/dev/null";
	open STDOUT, ">/dev/null";
	return 1;
}
sub prettylist {
	my $this=shift;
	my @objects=@_;
	return "nothing" if ! @objects;
	@objects = map { $_ == $this ? 'you' : $_->prettyname }
		grep ref, @objects;
	$objects[$#objects] = 'and '.$objects[$#objects] if @objects > 1;
	return join((@objects > 2) ? ', ' : ' ', @objects);
}
sub prettyname {
	my $a = $_[0]->article;
	$a.=" " if length $a;
	return $a.$_[0]->name;
}
sub untaint {
	($_[0]->[_USEPATH]) = $_[0]->[_USEPATH] =~ /(.*)/;
	($_[0]->[_ABSPATH]) = $_[0]->[_ABSPATH] =~ /(.*)/;
	return $_[0];
}
sub DESTROY {}
sub AUTOLOAD {
	(my $field = our $AUTOLOAD) =~ s/.*://;
	$field=~s/^_/./;
	my $this=shift;
	
	my $file=$this->fieldfile($field);
	if (defined $file && -f _ && -x _) {
		$this->debuglog(type => "call", field => $field, message => \@_) if $debugging;
		return $this->_runmethod($file, @_);
	}
	elsif (@_) {
		$this->debuglog(type => "write", field => $field, message => \@_) if $debugging;
		return $this->_writefield("$$this[0]/$field", @_);
	}
	elsif (defined $file && -e _) {
		$this->debuglog(type => "read", field => $field) if $debugging;
		return $this->_readfield($file);
	}
	return;
}
sub _readfield {
	my $this=shift;
	my $file=shift;
	
	if ((-d _ || -l _)) {
		if (-e "$file/.mooix") {
			return Mooix::Thing->get($file, $this->[_ABSPATH]);
		}
		else {
			return undef if -l $file && ! -d _ && ! -f _;
		}
	}
	
	my $fh;
	if (! open ($fh, $file)) {
		$this->croak($file);
	}
	if (wantarray) {
		my $sticky=-k _;
		my @ret = grep defined, map {
			chomp;
			if ($sticky) {
				$_ = Mooix::Thing->get($_, $this->[_ABSPATH]);
			}
			$_
		} <$fh>;
		close $fh;
		return @ret;
	}
	local $/=undef;
	my $ret=<$fh>;
	close $fh;
	local $/="\n";
	chomp $ret;
	return $ret;
}
sub _writefield {
	my $this=shift;
	my $file=shift;
	if ($this->[_USEPATH] ne '.') {
		my ($field)=$file=~m:/([^/]+)$:;
		if ($field eq 'setfield') {
			$this->croak("cannot setfield a field named setfield");
		}
		my @ret=$this->setfield($field, @_);
		if (($? >> 8) != 0) {
			$!="";
			$this->croak("setfield $field failed");
		}
	}
	else {
		if (@_ == 1 && ref $_[0] eq __PACKAGE__) {
			my $src=$_[0]->[_ABSPATH];
			my $tmp="$file-new.$$";
			symlink($src, $tmp) || $this->croak($tmp);
			rename($tmp, $file) || $this->croak($file);
		}
		else {
			if (-l $file) {
				unlink "$file" || $this->croak("unlink $file");
			}
			open (my $fh, ">", "$file") || $this->croak($file);
			print $fh join("\n", @_);
			close $fh;
			
			if (@_ > 1 && ! grep { ref $_[0] ne __PACKAGE__} @_) {
				chmod(01644, $file) || $this->croak("chmod $file");
			}
		}
	}
	return @_ if wantarray;
	return $_[0];
}
sub _runmethod {
	my $this=shift;
	my $file=shift;
	my ($child_wtr, $child_rdr);
	pipe(PARENT_RDR, $child_wtr);
	pipe($child_rdr, PARENT_WTR);
	my $pid;
	if ($pid = fork) {
		$SIG{PIPE}=sub {};
		
		close PARENT_RDR;
		close PARENT_WTR;
		
		select $child_wtr; $|=1; select STDOUT;
		foreach (@_) {
			if (! ref $_) {
				my $f=$_;
				chomp $f;
				$f =~ s/\\/\\\\/g;
				$f =~ s/\n/\\n/g;
				print $child_wtr "\"$f\"\n";
			}
			else {
				print $child_wtr "$_\n";
			}
		}
		close $child_wtr;
		
		if ($this->[_EXEC_NEXT]) {
			print <$child_rdr>;
			close $child_rdr;
			waitpid($pid, 0);
			exit $? >> 8;
		}
		my @ret = map {
			chomp;
			if (/^mooix/) {
				$_=Mooix::Thing->get($_)
			}
			elsif (/^"(.*)"$/) {
				$_=$1;
				s/(?<!\\)\\n/\n/g;
				s/\\\\/\\/g;
			}
			$_
		} <$child_rdr>;
		close $child_rdr;
		
		waitpid($pid, 0);
		
		if ($debugging && defined wantarray) {
			$this->debuglog(type => "return", field => $file,
			                message => \@ret, replace => 1,
					value => \@ret)
		}
		
		return @ret if wantarray;
		return $ret[0];
	}
	else {
		close $child_rdr;
		close $child_wtr;
	
		open(STDIN, "<&PARENT_RDR");
		open(STDOUT, ">&PARENT_WTR");
		
		chdir($this->[_ABSPATH]) || $this->croak("chdir");
		$file=~s/^\Q$$this[0]\E\///;
		if (exists $ENV{THIS}) {
			CORE::exec("./$file") || $this->croak("exec ./$file");
		}
		else {
			CORE::exec("runmeth", "./$file") || $this->croak("runmeth ./$file");
		}
	}
}
use overload
	'""' => sub { "mooix:".$_[0]->[_ABSPATH] },
	'<=>' => sub {
		my $a = shift;
		my $b = shift;
		$a->[_INDEX] cmp $b->[_INDEX]
	},
	'eq' => sub {
		my $a = shift;
		my $b = shift;
		$a->[_ABSPATH] eq $b->[_ABSPATH]
	};
1
